from dataclasses import dataclass
from typing import Optional

from pyckaxe import (
    Block,
    BlockState,
    Breadcrumb,
    JsonValue,
    NbtCompound,
    ResourceLocation,
    to_nbt_compound,
)

from mcblueprints.lib.resource.material.material import Material
from mcblueprints.lib.resource.material.types import MaterialOrLocation

__all__ = ("MaterialDeserializer",)


class MaterialDeserializationException(Exception):
    pass


class MalformedMaterial(MaterialDeserializationException):
    def __init__(self, message: str, raw_material: JsonValue, breadcrumb: Breadcrumb):
        self.raw_material: JsonValue = raw_material
        self.breadcrumb: Breadcrumb = breadcrumb
        super().__init__(message)


# @implements ResourceDeserializer[JsonValue, Material]
@dataclass
class MaterialDeserializer:
    # @implements ResourceDeserializer
    def __call__(
        self,
        raw: JsonValue,
        *,
        breadcrumb: Optional[Breadcrumb] = None,
    ) -> Material:
        return self.deserialize(raw, breadcrumb or Breadcrumb())

    def or_location(self, raw: JsonValue, breadcrumb: Breadcrumb) -> MaterialOrLocation:
        """Deserialize a `Material` or `MaterialLocation` from a raw value."""
        # A string is assumed to be a resource location.
        if isinstance(raw, str):
            return Material @ ResourceLocation.from_string(raw)
        # Anything else is assumed to be a serialized resource.
        return self(raw, breadcrumb=breadcrumb)

    def deserialize(self, raw_material: JsonValue, breadcrumb: Breadcrumb) -> Material:
        """Deserialize a `Material` from a raw value."""
        block = self.deserialize_block(raw_material, breadcrumb)
        return Material(block=block)

    def deserialize_block(self, raw_block: JsonValue, breadcrumb: Breadcrumb) -> Block:
        # A string is assumed to be a basic block.
        if isinstance(raw_block, str):
            return Block(name=raw_block)

        # Otherwise we ought to have a concrete definition...
        if not isinstance(raw_block, dict):
            raise MalformedMaterial(
                f"Malformed block, at `{breadcrumb}`", raw_block, breadcrumb
            )

        # name (required, non-nullable, no default)
        raw_name = raw_block.get("name")
        breadcrumb_name = breadcrumb.name
        if not raw_name:
            raise MalformedMaterial(
                f"Missing `name`, at `{breadcrumb_name}`", raw_block, breadcrumb_name
            )
        name = self.deserialize_name(raw_name, breadcrumb_name)

        # state (optional, nullable, defaults to null)
        state: Optional[BlockState] = None
        if (raw_state := raw_block.get("state")) is not None:
            state = self.deserialize_state(raw_state, breadcrumb.state)

        # data (optional, nullable, defaults to null)
        data: Optional[NbtCompound] = None
        if (raw_data := raw_block.get("data")) is not None:
            data = self.deserialize_data(raw_data, breadcrumb.data)

        return Block(
            name=name,
            state=state,
            data=data,
        )

    def deserialize_name(self, raw_name: JsonValue, breadcrumb: Breadcrumb) -> str:
        if not isinstance(raw_name, str):
            raise MalformedMaterial(
                f"Malformed `name`, at `{breadcrumb}`", raw_name, breadcrumb
            )
        return raw_name

    def deserialize_state(
        self, raw_state: JsonValue, breadcrumb: Breadcrumb
    ) -> BlockState:
        if not isinstance(raw_state, dict):
            raise MalformedMaterial(
                f"Malformed `state`, at `{breadcrumb}`", raw_state, breadcrumb
            )
        return BlockState(**raw_state)

    def deserialize_data(
        self, raw_data: JsonValue, breadcrumb: Breadcrumb
    ) -> NbtCompound:
        if not isinstance(raw_data, dict):
            raise MalformedMaterial(
                f"Malformed `data`, at `{breadcrumb}`", raw_data, breadcrumb
            )
        return to_nbt_compound(raw_data)



from dataclasses import dataclass
from typing import Optional

from pyckaxe import BlockMap, Position, ResolutionContext

from mcblueprints.lib.resource.blueprint.palette_entry.abc.blueprint_palette_entry import (
    BlueprintPaletteEntry,
)
from mcblueprints.lib.resource.blueprint.types import BlueprintOrLocation
from mcblueprints.lib.resource.filter.types import FilterOrLocation

__all__ = ("BlueprintBlueprintPaletteEntry",)


@dataclass
class BlueprintBlueprintPaletteEntry(BlueprintPaletteEntry):
    blueprint: BlueprintOrLocation
    offset: Position
    filter: Optional[FilterOrLocation] = None

    async def merge(
        self, ctx: ResolutionContext, block_map: BlockMap, position: Position
    ):
        # Resolve the child blueprint.
        child_blueprint = await ctx[self.blueprint]

        # Flatten the child blueprint into its own block map independently.
        child_block_map = await child_blueprint.flatten(ctx)

        # If a filter is present, apply it to the child block map before merging it.
        if self.filter is not None:
            filter = await ctx[self.filter]
            await filter.apply(ctx, child_block_map)

        # Merge the converted child block map into the parent block map.
        child_offset = position - self.offset - child_blueprint.anchor
        block_map.merge(child_block_map, child_offset)



from abc import ABC, abstractmethod
from dataclasses import dataclass

from pyckaxe import BlockMap, ResolutionContext

__all__ = ("FilterRule",)


@dataclass(frozen=True)
class FilterRule(ABC):
    @abstractmethod
    async def apply(self, ctx: ResolutionContext, block_map: BlockMap):
        """Apply the filter rule to `block_map`."""



from dataclasses import dataclass
from typing import Optional

from pyckaxe import Block, BlockState, NbtCompound, Resource

__all__ = ("Material",)


@dataclass
class Material(Resource):
    block: Block

    def __str__(self) -> str:
        return str(self.block)

    def __repr__(self) -> str:
        return str(self)

    @property
    def name(self) -> str:
        return self.block.name

    @property
    def state(self) -> Optional[BlockState]:
        return self.block.state

    @property
    def data(self) -> Optional[NbtCompound]:
        return self.block.data



from typing import Union

from pyckaxe import ClassifiedResourceLocation, ResourceProcessingContext

from mcblueprints.lib.resource.filter.filter import Filter

FilterLocation = ClassifiedResourceLocation[Filter]
FilterOrLocation = Union[Filter, FilterLocation]
FilterProcessingContext = ResourceProcessingContext[Filter]



from dataclasses import dataclass
from typing import List

from pyckaxe import Block, BlockMap, ResolutionContext

from mcblueprints.lib.resource.filter.rule.abc.filter_rule import FilterRule

__all__ = ("KeepBlocksFilterRule",)


@dataclass(frozen=True)
class KeepBlocksFilterRule(FilterRule):
    blocks: List[Block]

    async def apply(self, ctx: ResolutionContext, block_map: BlockMap):
        block_map.keep_blocks(self.blocks)



from dataclasses import dataclass

from pyckaxe import Block, BlockMap, Position, ResolutionContext

from mcblueprints.lib.resource.blueprint.palette_entry.abc.blueprint_palette_entry import (
    BlueprintPaletteEntry,
)

__all__ = ("BlockBlueprintPaletteEntry",)


@dataclass
class BlockBlueprintPaletteEntry(BlueprintPaletteEntry):
    block: Block

    async def merge(
        self, ctx: ResolutionContext, block_map: BlockMap, position: Position
    ):
        # Set the corresponding block in the block map.
        block_map[position] = self.block



from dataclasses import dataclass
from typing import List

from pyckaxe import BlockMap, ResolutionContext

from mcblueprints.lib.resource.filter.rule.abc.filter_rule import FilterRule
from mcblueprints.lib.resource.material.types import MaterialOrLocation

__all__ = ("KeepMaterialsFilterRule",)


@dataclass(frozen=True)
class KeepMaterialsFilterRule(FilterRule):
    materials: List[MaterialOrLocation]

    async def apply(self, ctx: ResolutionContext, block_map: BlockMap):
        materials = [await ctx[block] for block in self.materials]
        blocks = [material.block for material in materials]
        block_map.keep_blocks(blocks)



from typing import Any

__all__ = (
    "MaterialLocation",
    "MaterialOrLocation",
    "MaterialProcessingContext",
)


MaterialLocation = Any
MaterialOrLocation = Any
MaterialProcessingContext = Any



from typing import Union

from pyckaxe import ClassifiedResourceLocation, ResourceProcessingContext

from mcblueprints.lib.resource.material.material import Material

MaterialLocation = ClassifiedResourceLocation[Material]
MaterialOrLocation = Union[Material, MaterialLocation]
MaterialProcessingContext = ResourceProcessingContext[Material]



from typing import Union

from pyckaxe import ClassifiedResourceLocation, ResourceProcessingContext

from mcblueprints.lib.resource.blueprint.blueprint import Blueprint

BlueprintLocation = ClassifiedResourceLocation[Blueprint]
BlueprintOrLocation = Union[Blueprint, BlueprintLocation]
BlueprintProcessingContext = ResourceProcessingContext[Blueprint]



from pathlib import Path
from typing import Any

import click
from pyckaxe.cli.utils import asyncify
from pyckaxe.utils import LOG_LEVELS, setup_logging

from mcblueprints import __version__
from mcblueprints.build.blueprints_build_context import (
    DEFAULT_BLUEPRINT_CACHE_SIZE,
    DEFAULT_BLUEPRINTS_REGISTRY,
    DEFAULT_FILTER_CACHE_SIZE,
    DEFAULT_FILTERS_REGISTRY,
    DEFAULT_GENERATED_STRUCTURES_REGISTRY,
    DEFAULT_MATCH_FILES,
    DEFAULT_MATERIAL_CACHE_SIZE,
    DEFAULT_MATERIALS_REGISTRY,
    BlueprintsBuildContext,
)

__all__ = ("run",)


PROG_NAME = "mcblueprints"


@click.group()
@click.version_option(__version__, "-v", "--version")
@click.option(
    "-l",
    "--log",
    type=click.Choice(LOG_LEVELS, case_sensitive=False),
    default=LOG_LEVELS[2],
    help="The level of verbosity to log at.",
)
@click.option(
    "-ll",
    "--detailed-logs/--no-detailed-logs",
    default=False,
    help="Whether to use the detailed logging format.",
)
def cli(log: str, detailed_logs: bool):
    setup_logging(level=log.upper(), detailed=detailed_logs)


@cli.command(
    "build",
    help="Build Minecraft structures from mcblueprints.",
)
@click.option(
    "--input",
    "input_path",
    type=click.Path(exists=True, resolve_path=True),
    required=True,
    callback=lambda ctx, param, value: Path(value),
    help="The path to the data pack to read the input.",
)
@click.option(
    "--output",
    "output_path",
    type=click.Path(resolve_path=True),
    required=True,
    callback=lambda ctx, param, value: Path(value),
    help="The path to the data pack to dump the output.",
)
@click.option(
    "--data_version",
    "data_version",
    type=int,
    required=True,
    help="The data version to use in generated structures.",
)
@click.option(
    "--match_files",
    "match_files",
    type=str,
    help="The glob pattern to match files against."
    + f" Defaults to: {DEFAULT_MATCH_FILES}",
)
@click.option(
    "--blueprints_registry",
    "blueprints_registry",
    type=str,
    help="The registry where custom blueprints are located."
    + f" Defaults to: {DEFAULT_BLUEPRINTS_REGISTRY}",
)
@click.option(
    "--filters_registry",
    "filters_registry",
    type=str,
    help="The registry where custom filters are located."
    + f" Defaults to: {DEFAULT_FILTERS_REGISTRY}",
)
@click.option(
    "--materials_registry",
    "materials_registry",
    type=str,
    help="The registry where custom materials are located."
    + f" Defaults to: {DEFAULT_MATERIALS_REGISTRY}",
)
@click.option(
    "--blueprint_cache_size",
    type=int,
    help="The maximum number of blueprints to keep cached in memory."
    + "Set to 0 to disable caching. Set to -1 for an unbounded cache."
    + f" Defaults to: {DEFAULT_BLUEPRINT_CACHE_SIZE}",
)
@click.option(
    "--filter_cache_size",
    type=int,
    help="The maximum number of filters to keep cached in memory."
    + "Set to 0 to disable caching. Set to -1 for an unbounded cache."
    + f" Defaults to: {DEFAULT_FILTER_CACHE_SIZE}",
)
@click.option(
    "--material_cache_size",
    type=int,
    help="The maximum number of materials to keep cached in memory."
    + "Set to 0 to disable caching. Set to -1 for an unbounded cache."
    + f" Defaults to: {DEFAULT_MATERIAL_CACHE_SIZE}",
)
@click.option(
    "--generated_structures_registry",
    "generated_structures_registry",
    type=str,
    help="The registry where vanilla structures are located."
    + f" Defaults to: {DEFAULT_GENERATED_STRUCTURES_REGISTRY}",
)
@click.option(
    "--generated_namespace",
    "generated_namespace",
    type=str,
    help="A separate namespace to use for generated resources.",
)
@click.option(
    "--generated_prefix",
    "generated_prefix",
    type=str,
    help="A prefix to apply to the locations of generated resources.",
)
@asyncify
async def cli_build(**kwargs: Any):
    filtered_args = {k: v for k, v in kwargs.items() if v is not None}
    ctx = BlueprintsBuildContext(**filtered_args)
    await ctx.build()


def run():
    cli(prog_name=PROG_NAME)



import asyncio
from dataclasses import dataclass, field
from datetime import datetime
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, Optional, Set, Tuple, cast

from pyckaxe import (
    CommonResourceLocationResolver,
    CommonResourceResolver,
    CommonResourceScanner,
    JsonResourceLoader,
    LRUResourceCache,
    NbtResourceDumper,
    PhysicalPack,
    PhysicalRegistryLocation,
    ResourceCache,
    ResourceDumperSet,
    ResourceLocationResolverSet,
    ResourceProcessingContext,
    ResourceResolverSet,
    ResourceScannerSet,
    ResourceTransformerSet,
    StaticResourceCache,
    Structure,
    StructureSerializer,
    UnboundedResourceCache,
    WritablePack,
)
from pyckaxe.utils import walk_exception

from mcblueprints.lib import (
    Blueprint,
    BlueprintDeserializer,
    BlueprintTransformer,
    Filter,
    FilterDeserializer,
    Material,
    MaterialDeserializer,
)

__all__ = ("BlueprintsBuildContext",)


DEFAULT = cast(Any, ...)

DEFAULT_BLUEPRINTS_REGISTRY = "blueprints"
DEFAULT_FILTERS_REGISTRY = "filters"
DEFAULT_MATERIALS_REGISTRY = "materials"

DEFAULT_BLUEPRINT_CACHE_SIZE = 1000
DEFAULT_FILTER_CACHE_SIZE = 1000
DEFAULT_MATERIAL_CACHE_SIZE = 1000

DEFAULT_MATCH_FILES = "[!!]*"

DEFAULT_GENERATED_STRUCTURES_REGISTRY = "structures"


@dataclass
class BlueprintsBuildContext:
    input_path: Path
    output_path: Path

    data_version: int

    match_files: str = DEFAULT_MATCH_FILES

    blueprints_registry: str = DEFAULT_BLUEPRINTS_REGISTRY
    filters_registry: str = DEFAULT_FILTERS_REGISTRY
    materials_registry: str = DEFAULT_MATERIALS_REGISTRY

    blueprint_cache_size: int = DEFAULT_BLUEPRINT_CACHE_SIZE
    filter_cache_size: int = DEFAULT_FILTER_CACHE_SIZE
    material_cache_size: int = DEFAULT_MATERIAL_CACHE_SIZE

    generated_structures_registry: str = DEFAULT_GENERATED_STRUCTURES_REGISTRY

    generated_namespace: Optional[str] = None
    generated_prefix: Optional[str] = None

    log: Logger = field(default=DEFAULT)

    blueprint_cache: ResourceCache[Blueprint] = field(default=DEFAULT)
    filter_cache: ResourceCache[Filter] = field(default=DEFAULT)
    material_cache: ResourceCache[Material] = field(default=DEFAULT)

    blueprint_deserializer: BlueprintDeserializer = field(default=DEFAULT)
    filter_deserializer: FilterDeserializer = field(default=DEFAULT)
    material_deserializer: MaterialDeserializer = field(default=DEFAULT)

    material_loader: JsonResourceLoader[Material] = field(default=DEFAULT)
    filter_loader: JsonResourceLoader[Filter] = field(default=DEFAULT)
    blueprint_loader: JsonResourceLoader[Blueprint] = field(default=DEFAULT)

    blueprints_registry_parts: Tuple[str, ...] = field(init=False)
    filters_registry_parts: Tuple[str, ...] = field(init=False)
    materials_registry_parts: Tuple[str, ...] = field(init=False)
    generated_structures_registry_parts: Tuple[str, ...] = field(init=False)

    generated_prefix_parts: Optional[Tuple[str, ...]] = field(init=False)

    transformers: ResourceTransformerSet = field(
        init=False, default_factory=lambda: ResourceTransformerSet()
    )
    blueprint_transformer: BlueprintTransformer = field(default=DEFAULT)

    dumpers: ResourceDumperSet = field(
        init=False, default_factory=lambda: ResourceDumperSet()
    )
    structure_dumper: NbtResourceDumper[Structure] = field(default=DEFAULT)

    input_pack: PhysicalPack = field(init=False)

    build_errors: Set[Exception] = field(init=False, default_factory=set)

    def __str__(self) -> str:
        return self.output_path.name

    def __post_init__(self):
        # Make sure the output path is absolute.
        if not self.output_path.is_absolute():
            raise ValueError(
                f"Expected absolute output path, but got: {self.output_path}"
            )
        # Create a logger, if one does not already exist.
        if self.log is DEFAULT:
            self.log = getLogger(f"{self}")
        # Split paths into parts.
        self.blueprints_registry_parts = tuple(self.blueprints_registry.split("/"))
        self.filters_registry_parts = tuple(self.filters_registry.split("/"))
        self.materials_registry_parts = tuple(self.materials_registry.split("/"))
        self.generated_structures_registry_parts = tuple(
            self.generated_structures_registry.split("/")
        )
        self.generated_prefix_parts = (
            tuple(self.generated_prefix.split("/")) if self.generated_prefix else None
        )
        # Create caches that do not already exist.
        if self.blueprint_cache is DEFAULT:
            if self.blueprint_cache_size > 0:
                self.blueprint_cache = LRUResourceCache(size=self.blueprint_cache_size)
            elif self.blueprint_cache_size == 0:
                self.blueprint_cache = StaticResourceCache()
            else:
                self.blueprint_cache = UnboundedResourceCache()
        if self.filter_cache is DEFAULT:
            if self.filter_cache_size > 0:
                self.filter_cache = LRUResourceCache(size=self.filter_cache_size)
            elif self.filter_cache_size == 0:
                self.filter_cache = StaticResourceCache()
            else:
                self.filter_cache = UnboundedResourceCache()
        if self.material_cache is DEFAULT:
            if self.material_cache_size > 0:
                self.material_cache = LRUResourceCache(size=self.material_cache_size)
            elif self.material_cache_size == 0:
                self.material_cache = StaticResourceCache()
            else:
                self.material_cache = UnboundedResourceCache()
        # Create serializers that do no already exist.
        if self.material_deserializer is DEFAULT:
            self.material_deserializer = MaterialDeserializer()
        if self.filter_deserializer is DEFAULT:
            self.filter_deserializer = FilterDeserializer(
                material_deserializer=self.material_deserializer,
            )
        if self.blueprint_deserializer is DEFAULT:
            self.blueprint_deserializer = BlueprintDeserializer(
                filter_deserializer=self.filter_deserializer,
                material_deserializer=self.material_deserializer,
            )
        # Create loaders that do not already exist.
        if self.blueprint_loader is DEFAULT:
            self.blueprint_loader = JsonResourceLoader(self.blueprint_deserializer)
        if self.filter_loader is DEFAULT:
            self.filter_loader = JsonResourceLoader(self.filter_deserializer)
        if self.material_loader is DEFAULT:
            self.material_loader = JsonResourceLoader(self.material_deserializer)
        # Create and register transformers that do not already exist.
        if self.blueprint_transformer is DEFAULT:
            self.blueprint_transformer = BlueprintTransformer(
                generated_namespace=self.generated_namespace,
                generated_prefix_parts=self.generated_prefix_parts,
            )
        self.transformers[Blueprint] = self.blueprint_transformer
        # Create and register dumpers that do not already exist.
        if self.structure_dumper is DEFAULT:
            self.structure_dumper = NbtResourceDumper(
                serializer=StructureSerializer(data_version=self.data_version),
                options=dict(gzipped=True),
            )
        self.dumpers[Structure] = self.structure_dumper
        # Create a representation of the input pack.
        self.input_pack = PhysicalPack(self.input_path)

    async def build(self):
        self.log.info(f"Reading from: {self.input_path}")
        self.log.info(f"Writing to: {self.output_path}")

        # Collect blueprint registries.
        self.log.info(f"Collecting blueprint registries:")
        blueprint_registries = await self.input_pack.get_registries(
            *self.blueprints_registry_parts
        )
        for registry in blueprint_registries:
            self.log.info(f"  - {registry}")

        # Warn if no registries were found.
        if not blueprint_registries:
            self.log.error(
                f"No blueprint registries found matching: {self.blueprints_registry}"
            )
            return

        # Create a separate task for each registry.
        tasks = [self.build_registry(registry) for registry in blueprint_registries]

        # Start the build timer.
        time_start = datetime.utcnow()

        # Build registries asynchronously from one another.
        await asyncio.gather(*tasks)

        # Stop the build timer.
        time_elapsed = datetime.utcnow() - time_start

        # Let the user know that processing is finished and whether it was successful.
        if self.build_errors:
            self.log.error(
                f"Failed to finish building pack {self}"
                + f" with {len(self.build_errors)} errors:"
            )
            for i, error in enumerate(self.build_errors):
                prefix = f"  [{i + 1}] "
                pad = len(prefix) * " "
                self.log.error(f"{prefix}{error}")
                causes = list(walk_exception(error))
                for cause in causes[:-1]:
                    self.log.error(f"{pad}├ {cause}")
                self.log.error(f"{pad}└ {causes[-1]}")
        else:
            self.log.info(f"Finished building pack {self}")

        # Output some final metrics.
        self.log.info(
            "Number of blueprints cached: "
            + f"{len(self.blueprint_cache)}"
            + (
                f" / {self.blueprint_cache_size}"
                if self.blueprint_cache_size >= 0
                else ""
            )
        )
        self.log.info(
            "Number of blueprint filters cached: "
            + f"{len(self.filter_cache)}"
            + (f" / {self.filter_cache_size}" if self.filter_cache_size >= 0 else "")
        )
        self.log.info(
            "Number of materials cached: "
            + f"{len(self.material_cache)}"
            + (
                f" / {self.material_cache_size}"
                if self.material_cache_size >= 0
                else ""
            )
        )
        self.log.info(f"Time elapsed: {time_elapsed}")

    async def build_registry(self, blueprints_registry: PhysicalRegistryLocation):
        self.log.info(f"Processing registry {blueprints_registry}")
        # Create and register input scanners.
        scanners = ResourceScannerSet(
            CommonResourceScanner[Blueprint](blueprints_registry, Blueprint),
        )
        # Create input location resolvers.
        blueprint_location_resolver = CommonResourceLocationResolver(
            path=blueprints_registry.namespace.path.parent,
            parts=self.blueprints_registry_parts,
        )
        filter_location_resolver = CommonResourceLocationResolver(
            path=blueprints_registry.namespace.path.parent,
            parts=self.filters_registry_parts,
        )
        material_location_resolver = CommonResourceLocationResolver(
            path=blueprints_registry.namespace.path.parent,
            parts=self.materials_registry_parts,
        )
        # Create and register input resolvers.
        resolvers = ResourceResolverSet()
        resolvers[Blueprint] = CommonResourceResolver[Blueprint](
            location_resolver=blueprint_location_resolver,
            loader=self.blueprint_loader,
            cache=self.blueprint_cache,
        )
        resolvers[Filter] = CommonResourceResolver[Filter](
            location_resolver=filter_location_resolver,
            loader=self.filter_loader,
            cache=self.filter_cache,
        )
        resolvers[Material] = CommonResourceResolver[Material](
            location_resolver=material_location_resolver,
            loader=self.material_loader,
            cache=self.material_cache,
        )
        # Create and register output location resolvers.
        output_location_resolvers = ResourceLocationResolverSet()
        output_location_resolvers[Structure] = CommonResourceLocationResolver(
            path=Path(self.output_path / "data"),
            parts=self.generated_structures_registry_parts,
        )
        # Create a representation of the output pack.
        output_pack = WritablePack(
            path=self.output_path,
            location_resolvers=output_location_resolvers,
            dumpers=self.dumpers,
        )
        # Process and transform each type of resource.
        await self.process_all(scanners, resolvers, self.transformers, output_pack)
        # Print a message when we're all done with this registry.
        self.log.info(f"Finished processing registry {blueprints_registry}")

    async def process_all(
        self,
        scanners: ResourceScannerSet,
        resolvers: ResourceResolverSet,
        transformers: ResourceTransformerSet,
        output_pack: WritablePack,
    ):
        for scanner in scanners:
            self.log.info(f"Processing {scanner}")
            count_inputs = 0
            count_outputs = 0
            count_errors = 0
            async for input_location in scanner(match=self.match_files):
                self.log.debug(f"-> {input_location}")
                count_inputs += 1
                try:
                    input_resource = await resolvers(input_location)
                    ctx = ResourceProcessingContext[Any](
                        resolver_set=resolvers,
                        resource=input_resource,
                        location=input_location,
                    )
                    async for output_resource, output_location in transformers(ctx):
                        self.log.debug(
                            f"  -> {output_location} (#{id(output_resource)})"
                        )
                        await output_pack.dump(output_resource, output_location)
                        count_outputs += 1
                except Exception as ex:
                    self.log.exception(
                        f"Error while processing resource {input_location}"
                    )
                    self.build_errors.add(ex)
                    count_errors += 1
            self.log.info(
                f"Finished processing {scanner}"
                + (
                    f" with {count_inputs} inputs -> {count_outputs} outputs"
                    if count_inputs > 0
                    else " (nothing to process)"
                )
                + (f" (with {count_errors} errors)" if count_errors > 0 else "")
            )



from dataclasses import dataclass
from typing import AsyncIterable, Optional, Tuple, TypeVar

from pyckaxe import Namespace, Resource, ResourceLocation, Structure, StructureLocation

from mcblueprints.lib.resource.blueprint.palette_entry.abc.blueprint_palette_entry import (
    BlueprintPaletteEntry,
)
from mcblueprints.lib.resource.blueprint.types import BlueprintProcessingContext

__all__ = ("BlueprintTransformer",)


ResourceType = TypeVar("ResourceType", bound=Resource)
PaletteEntryType = TypeVar("PaletteEntryType", bound=BlueprintPaletteEntry)


# @implements ResourceTransformer[Blueprint]
@dataclass
class BlueprintTransformer:
    """
    Transforms a blueprint into a structure.

    Attributes
    ----------
    structure_serializer
        The structure serializer to use, including data version.
    generated_namespace
        A separate namespace to use for generated resources.
    generated_prefix_parts
        A prefix to apply to the locations of generated resources.
    """

    generated_namespace: Optional[str] = None
    generated_prefix_parts: Optional[Tuple[str, ...]] = None

    # @implements ResourceTransformer
    def __call__(
        self, ctx: BlueprintProcessingContext
    ) -> AsyncIterable[Tuple[Resource, ResourceLocation]]:
        return self.transform(ctx)

    async def transform(
        self, ctx: BlueprintProcessingContext
    ) -> AsyncIterable[Tuple[Resource, ResourceLocation]]:
        """Turn the blueprint into a structure NBT file."""
        structure = await ctx.resource.to_structure(ctx)
        structure_location = self.to_structure_location(ctx.location)
        yield structure, structure_location

    def to_structure_location(self, location: ResourceLocation) -> StructureLocation:
        # Map the blueprint location to a structure location.
        namespace = Namespace(name=self.generated_namespace or location.namespace.name)
        parts = location.parts
        if self.generated_prefix_parts:
            parts = (*self.generated_prefix_parts, *parts)
        structure_location = ResourceLocation(namespace, parts)
        return Structure @ structure_location



from typing import Any

__all__ = (
    "BlueprintLocation",
    "BlueprintOrLocation",
    "BlueprintProcessingContext",
)


BlueprintLocation = Any
BlueprintOrLocation = Any
BlueprintProcessingContext = Any



from dataclasses import dataclass, field
from typing import Callable, Dict, List, Optional

from pyckaxe import Breadcrumb, JsonValue, ResourceLocation

from mcblueprints.lib.resource.filter.filter import Filter, FilterRule
from mcblueprints.lib.resource.filter.rule.keep_blocks_filter_rule import (
    KeepBlocksFilterRule,
)
from mcblueprints.lib.resource.filter.rule.keep_materials_filter_rule import (
    KeepMaterialsFilterRule,
)
from mcblueprints.lib.resource.filter.rule.replace_blocks_filter_rule import (
    ReplaceBlocksFilterRule,
)
from mcblueprints.lib.resource.filter.rule.replace_materials_filter_rule import (
    ReplaceMaterialsFilterRule,
)
from mcblueprints.lib.resource.filter.types import FilterOrLocation
from mcblueprints.lib.resource.material.material_deserializer import (
    MaterialDeserializer,
)

__all__ = ("FilterDeserializer",)


class FilterDeserializationException(Exception):
    pass


class MalformedFilter(FilterDeserializationException):
    def __init__(self, message: str, raw_filter: JsonValue, breadcrumb: Breadcrumb):
        self.raw_filter: JsonValue = raw_filter
        self.breadcrumb: Breadcrumb = breadcrumb
        super().__init__(message)


class MalformedRule(FilterDeserializationException):
    def __init__(self, message: str, raw_rule: JsonValue, breadcrumb: Breadcrumb):
        self.raw_rule = raw_rule
        self.breadcrumb: Breadcrumb = breadcrumb
        super().__init__(message)


# @implements ResourceDeserializer[JsonValue, Filter]
@dataclass
class FilterDeserializer:
    material_deserializer: MaterialDeserializer

    rule_deserializers: Dict[
        str, Callable[[Dict[str, JsonValue], Breadcrumb], FilterRule]
    ] = field(init=False)

    def __post_init__(self):
        self.rule_deserializers = {
            "keep_blocks": self.deserialize_keep_blocks_rule,
            "keep_materials": self.deserialize_keep_materials_rule,
            "replace_blocks": self.deserialize_replace_blocks_rule,
            "replace_materials": self.deserialize_replace_materials_rule,
        }

    # @implements ResourceDeserializer
    def __call__(
        self,
        raw: JsonValue,
        *,
        breadcrumb: Optional[Breadcrumb] = None,
    ) -> Filter:
        return self.deserialize(raw, breadcrumb or Breadcrumb())

    def or_location(self, raw: JsonValue, breadcrumb: Breadcrumb) -> FilterOrLocation:
        """Deserialize a `Filter` or `FilterLocation` from a raw value."""
        # A string is assumed to be a resource location.
        if isinstance(raw, str):
            return Filter @ ResourceLocation.from_string(raw)
        # Anything else is assumed to be a serialized resource.
        return self(raw, breadcrumb=breadcrumb)

    def deserialize(self, raw_filter: JsonValue, breadcrumb: Breadcrumb) -> Filter:
        """Deserialize a `Filter` from a raw value."""
        if isinstance(raw_filter, dict):
            # rules (required, non-nullable, no default)
            raw_rules = raw_filter.get("rules")
            breadcrumb_rules = breadcrumb.rules
            if raw_rules is None:
                raise MalformedFilter(
                    f"Missing `rules`, at `{breadcrumb_rules}`",
                    raw_filter,
                    breadcrumb_rules,
                )
            rules = self.deserialize_rules(raw_rules, breadcrumb_rules)

        elif isinstance(raw_filter, list):
            rules = self.deserialize_rules(raw_filter, breadcrumb)

        else:
            raise MalformedFilter(
                f"Malformed filter, at `{breadcrumb}`", raw_filter, breadcrumb
            )

        filter = Filter(rules=rules)

        return filter

    def deserialize_rules(
        self, raw_rules: JsonValue, breadcrumb: Breadcrumb
    ) -> List[FilterRule]:
        if not isinstance(raw_rules, list):
            raise MalformedFilter(
                f"Malformed `rules`, at `{breadcrumb}`", raw_rules, breadcrumb
            )
        rules = [
            self.deserialize_rule(raw_rule, breadcrumb[i])
            for i, raw_rule in enumerate(raw_rules)
        ]
        return rules

    def deserialize_rule(
        self, raw_rule: JsonValue, breadcrumb: Breadcrumb
    ) -> FilterRule:
        if not isinstance(raw_rule, dict):
            raise MalformedRule(
                f"Malformed rule, at `{breadcrumb}`", raw_rule, breadcrumb
            )

        rule_type = raw_rule.get("type")
        breadcrumb_type = breadcrumb.type
        if rule_type is None:
            raise MalformedRule(
                f"Missing `type`, at `{breadcrumb_type}`", raw_rule, breadcrumb_type
            )
        if not isinstance(rule_type, str):
            raise MalformedRule(
                f"Malformed `type`, at `{breadcrumb_type}`", raw_rule, breadcrumb_type
            )

        rule_deserializer = self.rule_deserializers.get(rule_type)

        if not rule_deserializer:
            raise MalformedRule(
                f"Unknown type `{rule_type}`, at `{breadcrumb_type}`",
                raw_rule,
                breadcrumb,
            )

        rule = rule_deserializer(raw_rule, breadcrumb)

        return rule

    def deserialize_keep_blocks_rule(
        self, raw_rule: Dict[str, JsonValue], breadcrumb: Breadcrumb
    ) -> KeepBlocksFilterRule:
        # blocks (required, non-nullable, no default)
        raw_blocks = raw_rule.get("blocks")
        breadcrumb_blocks = breadcrumb.blocks
        if raw_blocks is None:
            raise MalformedRule(
                f"Missing `blocks`, at `{breadcrumb_blocks}`",
                raw_rule,
                breadcrumb_blocks,
            )
        if not isinstance(raw_blocks, list):
            raise MalformedRule(
                f"Malformed `blocks`, at `{breadcrumb_blocks}`",
                raw_rule,
                breadcrumb_blocks,
            )
        blocks = [
            self.material_deserializer.deserialize_block(
                raw_block, breadcrumb_blocks[i]
            )
            for i, raw_block in enumerate(raw_blocks)
        ]

        return KeepBlocksFilterRule(blocks=blocks)

    def deserialize_keep_materials_rule(
        self, raw_rule: Dict[str, JsonValue], breadcrumb: Breadcrumb
    ) -> KeepMaterialsFilterRule:
        # materials (required, non-nullable, no default)
        raw_materials = raw_rule.get("materials")
        breadcrumb_materials = breadcrumb.materials
        if raw_materials is None:
            raise MalformedRule(
                f"Missing `materials`, at `{breadcrumb_materials}`",
                raw_rule,
                breadcrumb_materials,
            )
        if not isinstance(raw_materials, list):
            raise MalformedRule(
                f"Malformed `materials`, at `{breadcrumb_materials}`",
                raw_rule,
                breadcrumb_materials,
            )
        materials = [
            self.material_deserializer.or_location(
                raw_material, breadcrumb_materials[i]
            )
            for i, raw_material in enumerate(raw_materials)
        ]

        return KeepMaterialsFilterRule(materials=materials)

    def deserialize_replace_blocks_rule(
        self, raw_rule: Dict[str, JsonValue], breadcrumb: Breadcrumb
    ) -> ReplaceBlocksFilterRule:
        # blocks (required, non-nullable, no default)
        raw_blocks = raw_rule.get("blocks")
        breadcrumb_blocks = breadcrumb.blocks
        if raw_blocks is None:
            raise MalformedRule(
                f"Missing `blocks`, at `{breadcrumb_blocks}`",
                raw_rule,
                breadcrumb_blocks,
            )
        if not isinstance(raw_blocks, list):
            raise MalformedRule(
                f"Malformed `blocks`, at `{breadcrumb_blocks}`",
                raw_rule,
                breadcrumb_blocks,
            )
        blocks = [
            self.material_deserializer.deserialize_block(
                raw_block, breadcrumb_blocks[i]
            )
            for i, raw_block in enumerate(raw_blocks)
        ]

        # replacement (required, non-nullable, no default)
        raw_replacement = raw_rule.get("replacement")
        breadcrumb_replacement = breadcrumb.replacement
        if raw_replacement is None:
            raise MalformedRule(
                f"Missing `replacement`, at `{breadcrumb_replacement}`",
                raw_rule,
                breadcrumb_replacement,
            )
        replacement = self.material_deserializer.deserialize_block(
            raw_replacement, breadcrumb_replacement
        )

        return ReplaceBlocksFilterRule(blocks=blocks, replacement=replacement)

    def deserialize_replace_materials_rule(
        self, raw_rule: Dict[str, JsonValue], breadcrumb: Breadcrumb
    ) -> ReplaceMaterialsFilterRule:
        # materials (required, non-nullable, no default)
        raw_materials = raw_rule.get("materials")
        breadcrumb_materials = breadcrumb.materials
        if raw_materials is None:
            raise MalformedRule(
                f"Missing `materials`, at `{breadcrumb_materials}`",
                raw_rule,
                breadcrumb_materials,
            )
        if not isinstance(raw_materials, list):
            raise MalformedRule(
                f"Malformed `materials`, at `{breadcrumb_materials}`",
                raw_rule,
                breadcrumb_materials,
            )
        materials = [
            self.material_deserializer.or_location(
                raw_material, breadcrumb_materials[i]
            )
            for i, raw_material in enumerate(raw_materials)
        ]

        # replacement (required, non-nullable, no default)
        raw_replacement = raw_rule.get("replacement")
        breadcrumb_replacement = breadcrumb.replacement
        if raw_replacement is None:
            raise MalformedRule(
                f"Missing `replacement`, at `{breadcrumb_replacement}`",
                raw_rule,
                breadcrumb_replacement,
            )
        replacement = self.material_deserializer.or_location(
            raw_replacement, breadcrumb_replacement
        )

        return ReplaceMaterialsFilterRule(materials=materials, replacement=replacement)



from dataclasses import dataclass

from pyckaxe import BlockMap, Position, ResolutionContext

from mcblueprints.lib.resource.blueprint.palette_entry.abc.blueprint_palette_entry import (
    BlueprintPaletteEntry,
)
from mcblueprints.lib.resource.material.types import MaterialOrLocation

__all__ = ("MaterialBlueprintPaletteEntry",)


@dataclass
class MaterialBlueprintPaletteEntry(BlueprintPaletteEntry):
    material: MaterialOrLocation

    async def merge(
        self, ctx: ResolutionContext, block_map: BlockMap, position: Position
    ):
        # Resolve the material.
        material = await ctx[self.material]
        # Set the corresponding block in the block map.
        block_map[position] = material.block



from dataclasses import dataclass
from typing import Dict, Iterable, List

from pyckaxe import BlockMap, Position, ResolutionContext, Resource, Structure

from mcblueprints.lib.resource.blueprint.palette_entry.abc.blueprint_palette_entry import (
    BlueprintPaletteEntry,
)

__all__ = (
    "BlueprintPalette",
    "BlueprintLayout",
    "Blueprint",
)


BlueprintPalette = Dict[str, BlueprintPaletteEntry]
BlueprintLayout = List[List[str]]


@dataclass
class Blueprint(Resource):
    size: Position
    anchor: Position
    palette: BlueprintPalette
    layout: BlueprintLayout

    def scan(self, symbol: str) -> Iterable[Position]:
        """Scan over the blueprint, looking for a particular symbol."""
        for y, floor in enumerate(self.layout):
            for x, row in enumerate(floor):
                yield from (
                    Position.from_xyz(x, y, z) for z, s in enumerate(row) if s == symbol
                )

    async def flatten(self, ctx: ResolutionContext) -> BlockMap:
        # Create a new block map to hold the final state.
        block_map = BlockMap(size=self.size)
        # Traverse palette entries in the order they are defined.
        for palette_key, palette_entry in self.palette.items():
            # Scan over the blueprint once per entry, looking for matching symbols.
            for offset in self.scan(palette_key):
                # Merge the palette entry into the block map at the offset.
                await palette_entry.merge(ctx, block_map, offset)
        return block_map

    async def to_structure(self, ctx: ResolutionContext) -> Structure:
        # Flatten the blueprint into a block map, and turn that into a structure.
        block_map = await self.flatten(ctx)
        structure = Structure.from_block_map(block_map)
        return structure



from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, cast

from pyckaxe import HERE, Block, Breadcrumb, JsonValue, Position, ResourceLocation

from mcblueprints.lib.resource.blueprint.blueprint import (
    Blueprint,
    BlueprintLayout,
    BlueprintPalette,
)
from mcblueprints.lib.resource.blueprint.palette_entry.abc.blueprint_palette_entry import (
    BlueprintPaletteEntry,
)
from mcblueprints.lib.resource.blueprint.palette_entry.block_blueprint_palette_entry import (
    BlockBlueprintPaletteEntry,
)
from mcblueprints.lib.resource.blueprint.palette_entry.blueprint_blueprint_palette_entry import (
    BlueprintBlueprintPaletteEntry,
)
from mcblueprints.lib.resource.blueprint.palette_entry.material_blueprint_palette_entry import (
    MaterialBlueprintPaletteEntry,
)
from mcblueprints.lib.resource.blueprint.palette_entry.void_blueprint_palette_entry import (
    VoidBlueprintPaletteEntry,
)
from mcblueprints.lib.resource.blueprint.types import BlueprintOrLocation
from mcblueprints.lib.resource.filter.filter_deserializer import FilterDeserializer
from mcblueprints.lib.resource.filter.types import FilterOrLocation
from mcblueprints.lib.resource.material.material import Material
from mcblueprints.lib.resource.material.material_deserializer import (
    MaterialDeserializer,
)

__all__ = ("BlueprintDeserializer",)


class BlueprintDeserializationException(Exception):
    pass


class MalformedBlueprint(BlueprintDeserializationException):
    def __init__(self, message: str, raw_blueprint: JsonValue, breadcrumb: Breadcrumb):
        self.raw_blueprint: JsonValue = raw_blueprint
        self.breadcrumb: Breadcrumb = breadcrumb
        super().__init__(message)


class MalformedPaletteEntry(BlueprintDeserializationException):
    def __init__(
        self, message: str, raw_palette_entry: JsonValue, breadcrumb: Breadcrumb
    ):
        self.raw_palette_entry: JsonValue = raw_palette_entry
        self.breadcrumb: Breadcrumb = breadcrumb
        super().__init__(message)


# @implements ResourceDeserializer[JsonValue, Blueprint]
@dataclass
class BlueprintDeserializer:
    filter_deserializer: FilterDeserializer
    material_deserializer: MaterialDeserializer

    palette_entry_deserializers: Dict[
        str, Callable[[str, Dict[str, JsonValue], Breadcrumb], BlueprintPaletteEntry]
    ] = field(init=False)

    def __post_init__(self):
        self.palette_entry_deserializers = {
            "block": self.deserialize_block_palette_entry,
            "blueprint": self.deserialize_blueprint_palette_entry,
            "material": self.deserialize_material_palette_entry,
            "void": self.deserialize_void_palette_entry,
        }

    # @implements ResourceDeserializer
    def __call__(
        self,
        raw: JsonValue,
        *,
        breadcrumb: Optional[Breadcrumb] = None,
    ) -> Blueprint:
        return self.deserialize(raw, breadcrumb or Breadcrumb())

    def or_location(
        self, raw: JsonValue, breadcrumb: Breadcrumb
    ) -> BlueprintOrLocation:
        """Deserialize a `Blueprint` or `BlueprintLocation` from a raw value."""
        # A string is assumed to be a resource location.
        if isinstance(raw, str):
            return Blueprint @ ResourceLocation.from_string(raw)
        # Anything else is assumed to be a serialized resource.
        return self(raw, breadcrumb=breadcrumb)

    def deserialize(
        self, raw_blueprint: JsonValue, breadcrumb: Breadcrumb
    ) -> Blueprint:
        """Deserialize a `Blueprint` from a raw value."""
        if not isinstance(raw_blueprint, dict):
            raise MalformedBlueprint(
                f"Malformed blueprint, at `{breadcrumb}`", raw_blueprint, breadcrumb
            )

        # size (required, non-nullable)
        raw_size = raw_blueprint.get("size")
        breadcrumb_size = breadcrumb.size
        if raw_size is None:
            raise MalformedBlueprint(
                f"Missing `size`, at `{breadcrumb_size}`",
                raw_blueprint,
                breadcrumb_size,
            )
        size = self.deserialize_size(raw_size, breadcrumb_size)

        # anchor (optional, non-nullable, has a default)
        anchor: Position = HERE
        if raw_anchor := raw_blueprint.get("anchor"):
            anchor = self.deserialize_anchor(raw_anchor, breadcrumb.anchor)

        # palette (required, non-nullable)
        raw_palette = raw_blueprint.get("palette")
        breadcrumb_palette = breadcrumb.palette
        if raw_palette is None:
            raise MalformedBlueprint(
                f"Missing `palette`, at `{breadcrumb_palette}`",
                raw_blueprint,
                breadcrumb_palette,
            )
        palette = self.deserialize_palette(raw_palette, breadcrumb_palette)

        # layout (required, non-nullable)
        raw_layout = raw_blueprint.get("layout")
        breadcrumb_layout = breadcrumb.layout
        if raw_layout is None:
            raise MalformedBlueprint(
                f"Missing `layout`, at `{breadcrumb_layout}`",
                raw_blueprint,
                breadcrumb_layout,
            )
        layout = self.deserialize_layout(raw_layout, breadcrumb_layout)

        blueprint = Blueprint(
            size=size,
            anchor=anchor,
            palette=palette,
            layout=layout,
        )

        return blueprint

    def deserialize_size(self, raw_size: JsonValue, breadcrumb: Breadcrumb) -> Position:
        if not isinstance(raw_size, list):
            raise MalformedBlueprint(
                f"Malformed `size`, at `{breadcrumb}`", raw_size, breadcrumb
            )
        return ~Position.from_list(cast(Any, raw_size))

    def deserialize_anchor(
        self, raw_anchor: JsonValue, breadcrumb: Breadcrumb
    ) -> Position:
        if not isinstance(raw_anchor, list):
            raise MalformedBlueprint(
                f"Malformed `anchor`, at `{breadcrumb}`", raw_anchor, breadcrumb
            )
        return ~Position.from_list(cast(Any, raw_anchor))

    def deserialize_palette(
        self, raw_palette: JsonValue, breadcrumb: Breadcrumb
    ) -> BlueprintPalette:
        if not isinstance(raw_palette, dict):
            raise MalformedBlueprint(
                f"Malformed `palette`, at `{breadcrumb}`", raw_palette, breadcrumb
            )
        palette: Dict[str, BlueprintPaletteEntry] = {}
        for i, (palette_key, raw_palette_entry) in enumerate(raw_palette.items()):
            if len(palette_key) != 1:
                raise MalformedPaletteEntry(
                    f"Palette key `{palette_key}` is not a single character,"
                    + f" at `{breadcrumb[palette_key]}`",
                    raw_palette_entry,
                    breadcrumb[palette_key],
                )
            palette[palette_key] = self.deserialize_palette_entry(
                palette_key, raw_palette_entry, breadcrumb[i]
            )
        return palette

    def deserialize_palette_entry(
        self, palette_key: str, raw_palette_entry: JsonValue, breadcrumb: Breadcrumb
    ) -> BlueprintPaletteEntry:
        # A string is assumed to be a basic block.
        if isinstance(raw_palette_entry, str):
            material = Material(block=Block(name=raw_palette_entry))
            return MaterialBlueprintPaletteEntry(key=palette_key, material=material)

        # Otherwise we ought to have a concrete definition...

        if not isinstance(raw_palette_entry, dict):
            raise MalformedPaletteEntry(
                f"Malformed palette entry, at `{breadcrumb}`",
                raw_palette_entry,
                breadcrumb,
            )

        palette_entry_type = raw_palette_entry.get("type")
        breadcrumb_type = breadcrumb.type
        if palette_entry_type is None:
            raise MalformedPaletteEntry(
                f"Missing `type`, at `{breadcrumb_type}`",
                raw_palette_entry,
                breadcrumb_type,
            )
        if not isinstance(palette_entry_type, str):
            raise MalformedPaletteEntry(
                f"Malformed `type`, at `{breadcrumb_type}`",
                raw_palette_entry,
                breadcrumb_type,
            )

        palette_entry_deserializer = self.palette_entry_deserializers.get(
            palette_entry_type
        )

        if not palette_entry_deserializer:
            raise MalformedPaletteEntry(
                f"Unknown type `{palette_entry_type}`, at `{breadcrumb_type}`",
                raw_palette_entry,
                breadcrumb,
            )

        palette_entry = palette_entry_deserializer(
            palette_key, raw_palette_entry, breadcrumb
        )

        return palette_entry

    def deserialize_block_palette_entry(
        self,
        palette_key: str,
        raw_palette_entry: Dict[str, JsonValue],
        breadcrumb: Breadcrumb,
    ) -> BlockBlueprintPaletteEntry:
        block = self.material_deserializer.deserialize_block(
            raw_palette_entry, breadcrumb
        )
        return BlockBlueprintPaletteEntry(key=palette_key, block=block)

    def deserialize_blueprint_palette_entry(
        self,
        palette_key: str,
        raw_palette_entry: Dict[str, JsonValue],
        breadcrumb: Breadcrumb,
    ) -> BlueprintBlueprintPaletteEntry:
        # blueprint (required, non-nullable)
        raw_blueprint = raw_palette_entry.get("blueprint")
        breadcrumb_blueprint = breadcrumb.blueprint
        if raw_blueprint is None:
            raise MalformedPaletteEntry(
                f"Missing `blueprint`, at `{breadcrumb_blueprint}`",
                raw_palette_entry,
                breadcrumb_blueprint,
            )
        blueprint = self.or_location(raw_blueprint, breadcrumb_blueprint)

        # offset (optional, non-nullable, has a default)
        offset: Position = HERE
        if raw_offset := raw_palette_entry.get("offset"):
            if not isinstance(raw_offset, list):
                raise MalformedBlueprint(
                    f"Malformed `offset`, at `{breadcrumb.offset}`",
                    raw_offset,
                    breadcrumb.offset,
                )
            offset = ~Position.from_list(cast(Any, raw_offset))

        # filter (optional, nullable, defaults to null)
        filter: Optional[FilterOrLocation] = None
        if raw_filter := raw_palette_entry.get("filter"):
            filter = self.filter_deserializer.or_location(raw_filter, breadcrumb.filter)

        return BlueprintBlueprintPaletteEntry(
            key=palette_key, blueprint=blueprint, offset=offset, filter=filter
        )

    def deserialize_material_palette_entry(
        self,
        palette_key: str,
        raw_palette_entry: Dict[str, JsonValue],
        breadcrumb: Breadcrumb,
    ) -> MaterialBlueprintPaletteEntry:
        # material (required, non-nullable)
        raw_material = raw_palette_entry.get("material")
        if raw_material is None:
            raise MalformedPaletteEntry(
                f"Missing `material`, at `{breadcrumb.material}`",
                raw_palette_entry,
                breadcrumb.material,
            )
        material = self.material_deserializer.or_location(raw_material, breadcrumb)

        return MaterialBlueprintPaletteEntry(key=palette_key, material=material)

    def deserialize_void_palette_entry(
        self,
        palette_key: str,
        raw_palette_entry: Dict[str, JsonValue],
        breadcrumb: Breadcrumb,
    ) -> VoidBlueprintPaletteEntry:
        return VoidBlueprintPaletteEntry(key=palette_key)

    def deserialize_layout(
        self, raw_layout: JsonValue, breadcrumb: Breadcrumb
    ) -> BlueprintLayout:
        if not isinstance(raw_layout, list):
            raise MalformedBlueprint(
                f"Malformed `layout`, at `{breadcrumb}`", raw_layout, breadcrumb
            )

        layout: BlueprintLayout = []

        # Read the layout upside-down.
        for i, raw_layer in enumerate(reversed(raw_layout)):
            if raw_layer is None:
                layout.append([])
                continue

            if isinstance(raw_layer, str):
                raw_layer = list(raw_layer)

            if not isinstance(raw_layer, list):
                raise MalformedBlueprint(
                    f"Malformed `layout` layer, at `{breadcrumb[i]}`",
                    raw_layer,
                    breadcrumb[i],
                )

            layer: List[str] = []

            for j, raw_row in enumerate(raw_layer):
                if raw_row is None:
                    layer.append("")
                    continue

                if not isinstance(raw_row, str):
                    raise MalformedBlueprint(
                        f"Malformed `layout` row, at `{breadcrumb[i][j]}`",
                        raw_row,
                        breadcrumb[i][j],
                    )

                layer.append(raw_row)

            layout.append(layer)

        return layout



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mcblueprints',
 'mcblueprints.build',
 'mcblueprints.cli',
 'mcblueprints.lib',
 'mcblueprints.lib.resource',
 'mcblueprints.lib.resource.blueprint',
 'mcblueprints.lib.resource.blueprint.palette_entry',
 'mcblueprints.lib.resource.blueprint.palette_entry.abc',
 'mcblueprints.lib.resource.filter',
 'mcblueprints.lib.resource.filter.rule',
 'mcblueprints.lib.resource.filter.rule.abc',
 'mcblueprints.lib.resource.material']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0', 'nbtlib>=1.12.0,<2.0.0', 'pyckaxe>=0.2.0,<0.3.0']

extras_require = \
{'colors': ['colorama>=0.4.3,<0.5.0', 'colorlog>=4.2.1,<5.0.0'],
 'yaml': ['pyyaml>=5.3.1,<6.0.0']}

setup_kwargs = {
    'name': 'mcblueprints',
    'version': '0.1.0',
    'description': 'Text-based structures for Minecraft.',
    'long_description': '![logo]\n\n# Blueprints\n\n> Text-based structures for Minecraft.\n\n[![package-badge]](https://pypi.python.org/pypi/mcblueprints/)\n[![version-badge]](https://pypi.python.org/pypi/mcblueprints/)\n[![style-badge]](https://github.com/ambv/black)\n\nBlueprints are a text-based structure format for Minecraft optimized for human-readability. A blueprint compiles-down to a single NBT structure file that can be loaded with a structure block.\n\nHere are some reasons you may want to use blueprints:\n\n- They can be diff\'d and _properly_ included in version control, unlike their NBT equivalent.\n- They are far smaller and less repetitive than their SNBT equivalent, which is often used for version control.\n- They can be expressed purely through text, without having to think about the underlying structure format (and without having to open the game).\n- They can be updated to (and optimized for) a newer version of the game, just by re-running the CLI with the `--data_version` argument.\n- They have the potential to take up significantly less space than their NBT equivalent, for codebases that take advantage of composition.\n\nKeep in mind that - although they are optimized for human-readbility - blueprints aren\'t nearly as "hands-on" as editing structures in-game. There are pros and cons to either approach, and the case for blueprints should be weighed carefully based on project size and complexity.\n\n## Usage\n\nBlueprints are created and maintained in the same way as vanilla resources, but under a made-up `blueprints` folder.\n\nSee the [examples](#examples) and the [demo pack](https://github.com/Arcensoth/blueprints/tree/main/tests/datapacks/demo-datapack/data) for reference.\n\nThe most basic invocation of the CLI looks like this:\n\n```bash\npython -m blueprints build --input path/to/input/pack --output path/to/output/pack --data_version 2715\n```\n\n- `--input` is the path to the input pack. This is where your blueprints reside.\n- `--output` is the path to the output pack. This is where the generated structures files will be placed. This can be the same as the input pack, but beware of overwriting existing files.\n- `--data_version` is required and `2715` should be replaced with the [version of the game](https://minecraft.fandom.com/wiki/Data_version#List_of_data_versions) you are targeting.\n\nRun `python -m mcblueprints build --help` for a complete list of options.\n\n## Examples\n\nAll examples use YAML instead of JSON, but the YAML used is 1:1 convertible to/from JSON.\n\nSee the full [demo pack](https://github.com/Arcensoth/blueprints/tree/main/tests/datapacks/demo-datapack/data) for a complete set of examples.\n\nThis first example uses basic blocks to create a simple structure.\n\n![image](https://user-images.githubusercontent.com/1885643/118862799-20da5f80-b8ac-11eb-9ad3-23b50f251e32.png)\n\n**`fleecy_box:base`**\n\n[`data/fleecy_box/blueprints/base.yaml`](https://github.com/Arcensoth/blueprints/blob/main/tests/datapacks/demo-datapack/data/fleecy_box/blueprints/base.yaml)\n\n```yaml\n# Restrict the size of the structure. An error will be raised if anything extends\n# outside of these bounds. This goes by (x, y, z) or (length x height x width).\nsize: [5, 5, 5]\n\n# The palette maps characters to different types of palette entries that describe how to\n# populate the structure. These can be blocks as well as other blueprints.\npalette:\n  # Strings are assumed to be basic blocks.\n  _: minecraft:air\n  g: minecraft:glass\n  c: minecraft:glowstone\n  b: minecraft:bricks\n  # To define block states, use the block type entry.\n  P:\n    type: block\n    name: minecraft:quartz_pillar\n    state:\n      axis: y\n  X:\n    type: block\n    name: minecraft:tnt\n    state:\n      unstable: true\n  # To define NBT data, use the block type entry.\n  T:\n    type: block\n    name: minecraft:trapped_chest\n    data:\n      Items:\n        - id: minecraft:diamond\n          Count: 1b\n          Slot: 13b\n\n# The layout is a 2-D list of strings (a 3-D list of characters) that says how to build\n# the structure, piece by piece, using the palette. Note that the first section of the\n# layout corresponds to the top-most layer of blocks in the structure.\nlayout:\n  - - ggggg\n    - ggggg\n    - ggggg\n    - ggggg\n    - ggggg\n\n  - - ggggg\n    - g___g\n    - g___g\n    - g___g\n    - ggggg\n\n  - - ggggg\n    - g___g\n    - g_T_g\n    - g___g\n    - ggggg\n\n  - - cgggc\n    - g___g\n    - g_P_g\n    - g___g\n    - cgggc\n\n  - - bbbbb\n    - bbbbb\n    - bbXbb\n    - bbbbb\n    - bbbbb\n```\n\nThis next example uses composition and a filter to include a modified version of another blueprint.\n\n![image](https://user-images.githubusercontent.com/1885643/118862891-3c456a80-b8ac-11eb-94a7-763484c12069.png)\n\n**`fleecy_box:copper`**\n\n[`data/fleecy_box/blueprints/copper.yaml`](https://github.com/Arcensoth/blueprints/blob/main/tests/datapacks/demo-datapack/data/fleecy_box/blueprints/copper.yaml)\n\n```yaml\n# Make sure to account for any included structures in the final structure.\nsize: [5, 5, 5]\n\npalette:\n  # Blueprints are composable. They can be included within one another, and the final\n  # structure will be a flattened version with a minimal palette.\n  B:\n    type: blueprint\n    blueprint: fleecy_box:base\n    # A filter changes the way blocks are included from other blueprints.\n    filter: fleecy_box:copperize\n\nlayout:\n  - - B\n```\n\nNote the use of a filter: these can be used to change the way blocks are included from other blueprints.\n\n**`fleecy_box:copperize`**\n\n[`data/fleecy_box/filters/copperize.yaml`](https://github.com/Arcensoth/blueprints/blob/main/tests/datapacks/demo-datapack/data/fleecy_box/filters/copperize.yaml)\n\n```yaml\n# Replace one block with another block.\n- type: replace_blocks\n  blocks:\n    - minecraft:bricks\n  replacement: minecraft:cut_copper\n\n- type: replace_blocks\n  blocks:\n    - minecraft:glowstone\n    - minecraft:quartz_pillar\n  replacement: minecraft:copper_block\n\n# Keep only the listed blocks, discarding the rest.\n- type: keep_blocks\n  blocks:\n    - minecraft:copper_block\n    - minecraft:cut_copper\n```\n\n[logo]: ./logo.png\n[package-badge]: https://img.shields.io/pypi/v/blueprints.svg\n[version-badge]: https://img.shields.io/pypi/pyversions/blueprints.svg\n[style-badge]: https://img.shields.io/badge/code%20style-black-000000.svg\n',
    'author': 'Arcensoth',
    'author_email': 'arcensoth@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Arcensoth/blueprints',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import List

from pyckaxe import BlockMap, ResolutionContext, Resource

from mcblueprints.lib.resource.filter.rule.abc.filter_rule import FilterRule

__all__ = ("Filter",)


@dataclass
class Filter(Resource):
    rules: List[FilterRule]

    async def apply(self, ctx: ResolutionContext, block_map: BlockMap):
        # Apply all rules, in order.
        for rule in self.rules:
            await rule.apply(ctx, block_map)



from abc import ABC, abstractmethod
from dataclasses import dataclass

from pyckaxe import BlockMap, Position, ResolutionContext

__all__ = ("BlueprintPaletteEntry",)


@dataclass
class BlueprintPaletteEntry(ABC):
    key: str

    @abstractmethod
    async def merge(
        self, ctx: ResolutionContext, block_map: BlockMap, position: Position
    ):
        """Merge into `block_map` at `position`."""
